#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#ifdef _WIN32
#include <windows.h>
#include <process.h>
#else
#include <sys/mman.h>
#include <unistd.h>
#endif

#include "memmap.h"

struct mappedRegionS {
    void * addr;
    int length;
};

int memmap( const char *path, int id, int length, mappedRegion *hReg)
{
    struct stat stat_buf;
    int res;
    unsigned int i;
    int exists = 1;
    char buffer[1024];
#ifdef _WIN32
    PSECURITY_DESCRIPTOR pSD;
    HANDLE hFile, hMem;
    SECURITY_ATTRIBUTES  sa;
    DWORD sz;
#else
    int fdes;
#endif

    mappedRegion reg = (mappedRegion)malloc( sizeof( struct mappedRegionS));
    if (NULL == reg) return -3;
    *hReg = reg;

    /* check if file already exists and determine its length */
    res = stat( path, &stat_buf);
    if (res < 0) {
        if (errno == ENOENT) exists = 0;
        else return -1;
    }
    if (exists) reg->length = stat_buf.st_size;
    else reg->length = length;

#ifdef _WIN32
    /* create security descriptor (needed for Windows NT) */
    pSD = (PSECURITY_DESCRIPTOR) malloc( SECURITY_DESCRIPTOR_MIN_LENGTH );
    if( pSD == NULL ) return -2;

    InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl(pSD, TRUE, (PACL) NULL, FALSE);

    sa.nLength = sizeof(sa);
    sa.lpSecurityDescriptor = pSD;
    sa.bInheritHandle = TRUE;  

    /* create or open file */
    if (exists) {
        hFile = CreateFile ( path, GENERIC_READ | GENERIC_WRITE, 
            FILE_SHARE_READ | FILE_SHARE_WRITE, &sa, OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL, NULL);  
    }
    else {
        hFile = CreateFile ( path, GENERIC_READ | GENERIC_WRITE, 
            FILE_SHARE_READ | FILE_SHARE_WRITE, &sa, OPEN_ALWAYS,
            FILE_ATTRIBUTE_NORMAL, NULL);  
    }
    if (hFile == INVALID_HANDLE_VALUE) {   
        free( pSD);
        return -3;
    }
    if (! exists) {
        /* ensure that file is long enough and filled with zero */
        memset( buffer, 0, sizeof(buffer));
        for (i = 0; i < reg->length/sizeof(buffer); ++i) {
            if (! WriteFile( hFile, buffer, sizeof(buffer), &sz, NULL)) {
                return -3;
            }
        }
        if (! WriteFile( hFile, buffer, reg->length, &sz, NULL)) {
            return -3;
        }
    }
        
    /* create file mapping */
    sprintf( buffer, "%d", id);
    hMem = CreateFileMapping( hFile, &sa, PAGE_READWRITE, 0, 
            reg->length, buffer);
    free( pSD);
    if (NULL == hMem) return -3;

    /* map the file to addregs */
    reg->addr = MapViewOfFile( hMem, FILE_MAP_ALL_ACCESS, 0, 0, 0);
    if (NULL == reg->addr) return -3;

    CloseHandle( hFile);
    CloseHandle( hMem);

#else
    /* UNIX */
    if (exists) {
        /* open mapped file */
        fdes = open( path, O_RDWR, S_IRUSR | S_IWUSR);
        if (fdes < -1) return -1;
    }
    else /* not exists */ {
        /* create mapped file */
        fdes = open( path, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
        if (fdes < -1) return -1;
        /* ensure that file is long enough and filled with zero */
        memset( buffer, 0, sizeof(buffer));
        for (i = 0; i < reg->length/sizeof(buffer); ++i) {
            if (write( fdes, buffer, sizeof(buffer)) != sizeof(buffer)) {
                return -1;
            }
        }
        if (write( fdes, buffer, reg->length) != reg->length) {
            return -1;
        }
    }

    /* map the file to m_addr */
    reg->addr = mmap( NULL, reg->length,
        PROT_READ | PROT_WRITE, MAP_SHARED, fdes, 0);  
    close( fdes);
    if (reg->addr == (void *)-1) return -1;
#endif

    return 0;
}

void memunmap( mappedRegion *hReg)
{
    mappedRegion reg = *hReg;
    if (reg) {
#ifdef _WIN32
        if (reg->addr) {
            UnmapViewOfFile( reg->addr);
        }
#else
        if (reg->addr) {
            munmap( reg->addr, reg->length);
        }
#endif
        free( reg); 
    }
    *hReg = 0;
}

char* memGetAddr( mappedRegion *hReg)
{
    return (char *)((*hReg)->addr);
}

int memGetLength( mappedRegion *hReg)
{
    return (*hReg)->length;
}
